package xyz.lychee.lagfixer.libs.adventure.text;

import java.util.Iterator;
import java.util.Objects;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;
import xyz.lychee.lagfixer.libs.adventure.internal.Internals;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style;
import xyz.lychee.lagfixer.libs.annotations.Contract;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.examination.ExaminableProperty;

final class JoinConfigurationImpl implements JoinConfiguration {
   static final Function<ComponentLike, Component> DEFAULT_CONVERTOR = ComponentLike::asComponent;
   static final Predicate<ComponentLike> DEFAULT_PREDICATE = (componentLike) -> {
      return true;
   };
   static final JoinConfigurationImpl NULL = new JoinConfigurationImpl();
   static final JoinConfiguration STANDARD_NEW_LINES = JoinConfiguration.separator(Component.newline());
   static final JoinConfiguration STANDARD_SPACES = JoinConfiguration.separator(Component.space());
   static final JoinConfiguration STANDARD_COMMA_SEPARATED = JoinConfiguration.separator(Component.text(","));
   static final JoinConfiguration STANDARD_COMMA_SPACE_SEPARATED = JoinConfiguration.separator(Component.text(", "));
   static final JoinConfiguration STANDARD_ARRAY_LIKE = (JoinConfiguration)JoinConfiguration.builder().separator(Component.text(", ")).prefix(Component.text("[")).suffix(Component.text("]")).build();
   private final Component prefix;
   private final Component suffix;
   private final Component separator;
   private final Component lastSeparator;
   private final Component lastSeparatorIfSerial;
   private final Function<ComponentLike, Component> convertor;
   private final Predicate<ComponentLike> predicate;
   private final Style rootStyle;

   private JoinConfigurationImpl() {
      this.prefix = null;
      this.suffix = null;
      this.separator = null;
      this.lastSeparator = null;
      this.lastSeparatorIfSerial = null;
      this.convertor = DEFAULT_CONVERTOR;
      this.predicate = DEFAULT_PREDICATE;
      this.rootStyle = Style.empty();
   }

   private JoinConfigurationImpl(@NotNull final @NotNull BuilderImpl builder) {
      this.prefix = ComponentLike.unbox(builder.prefix);
      this.suffix = ComponentLike.unbox(builder.suffix);
      this.separator = ComponentLike.unbox(builder.separator);
      this.lastSeparator = ComponentLike.unbox(builder.lastSeparator);
      this.lastSeparatorIfSerial = ComponentLike.unbox(builder.lastSeparatorIfSerial);
      this.convertor = builder.convertor;
      this.predicate = builder.predicate;
      this.rootStyle = builder.rootStyle;
   }

   public @Nullable Component prefix() {
      return this.prefix;
   }

   public @Nullable Component suffix() {
      return this.suffix;
   }

   public @Nullable Component separator() {
      return this.separator;
   }

   public @Nullable Component lastSeparator() {
      return this.lastSeparator;
   }

   public @Nullable Component lastSeparatorIfSerial() {
      return this.lastSeparatorIfSerial;
   }

   public @NotNull Function<ComponentLike, Component> convertor() {
      return this.convertor;
   }

   public @NotNull Predicate<ComponentLike> predicate() {
      return this.predicate;
   }

   public @NotNull Style parentStyle() {
      return this.rootStyle;
   }

   public JoinConfiguration.@NotNull Builder toBuilder() {
      return new BuilderImpl(this);
   }

   public @NotNull Stream<? extends ExaminableProperty> examinableProperties() {
      return Stream.of(ExaminableProperty.of("prefix", this.prefix), ExaminableProperty.of("suffix", this.suffix), ExaminableProperty.of("separator", this.separator), ExaminableProperty.of("lastSeparator", this.lastSeparator), ExaminableProperty.of("lastSeparatorIfSerial", this.lastSeparatorIfSerial), ExaminableProperty.of("convertor", this.convertor), ExaminableProperty.of("predicate", this.predicate), ExaminableProperty.of("rootStyle", this.rootStyle));
   }

   public String toString() {
      return Internals.toString(this);
   }

   @Contract(
      pure = true
   )
   static @NotNull Component join(final @NotNull JoinConfiguration config, final @NotNull Iterable<? extends ComponentLike> components) {
      Objects.requireNonNull(config, "config");
      Objects.requireNonNull(components, "components");
      Iterator<? extends ComponentLike> it = components.iterator();
      if (!it.hasNext()) {
         return singleElementJoin(config, (ComponentLike)null);
      } else {
         ComponentLike component = (ComponentLike)Objects.requireNonNull((ComponentLike)it.next(), "Null elements in \"components\" are not allowed");
         int componentsSeen = 0;
         if (!it.hasNext()) {
            return singleElementJoin(config, component);
         } else {
            Component prefix = config.prefix();
            Component suffix = config.suffix();
            Function<ComponentLike, Component> convertor = config.convertor();
            Predicate<ComponentLike> predicate = config.predicate();
            Style rootStyle = config.parentStyle();
            boolean hasRootStyle = rootStyle != Style.empty();
            Component separator = config.separator();
            boolean hasSeparator = separator != null;
            TextComponent.Builder builder = hasRootStyle ? (TextComponent.Builder)Component.text().style(rootStyle) : Component.text();
            if (prefix != null) {
               builder.append(prefix);
            }

            while(component != null) {
               if (!predicate.test(component)) {
                  if (!it.hasNext()) {
                     break;
                  }

                  component = (ComponentLike)it.next();
               } else {
                  builder.append((Component)Objects.requireNonNull((Component)convertor.apply(component), "Null output from \"convertor\" is not allowed"));
                  ++componentsSeen;
                  if (!it.hasNext()) {
                     component = null;
                  } else {
                     component = (ComponentLike)Objects.requireNonNull((ComponentLike)it.next(), "Null elements in \"components\" are not allowed");
                     if (it.hasNext()) {
                        if (hasSeparator) {
                           builder.append(separator);
                        }
                     } else {
                        Component lastSeparator = null;
                        if (componentsSeen > 1) {
                           lastSeparator = config.lastSeparatorIfSerial();
                        }

                        if (lastSeparator == null) {
                           lastSeparator = config.lastSeparator();
                        }

                        if (lastSeparator == null) {
                           lastSeparator = config.separator();
                        }

                        if (lastSeparator != null) {
                           builder.append(lastSeparator);
                        }
                     }
                  }
               }
            }

            if (suffix != null) {
               builder.append(suffix);
            }

            return builder.build();
         }
      }
   }

   static @NotNull Component singleElementJoin(final @NotNull JoinConfiguration config, final @Nullable ComponentLike component) {
      Component prefix = config.prefix();
      Component suffix = config.suffix();
      Function<ComponentLike, Component> convertor = config.convertor();
      Predicate<ComponentLike> predicate = config.predicate();
      Style rootStyle = config.parentStyle();
      boolean hasRootStyle = rootStyle != Style.empty();
      if (prefix == null && suffix == null) {
         Object result;
         if (component != null && predicate.test(component)) {
            result = (Component)convertor.apply(component);
         } else {
            result = Component.empty();
         }

         return (Component)(hasRootStyle ? ((TextComponent.Builder)((TextComponent.Builder)Component.text().style(rootStyle)).append((Component)result)).build() : result);
      } else {
         TextComponent.Builder builder = Component.text();
         if (prefix != null) {
            builder.append(prefix);
         }

         if (component != null && predicate.test(component)) {
            builder.append((Component)convertor.apply(component));
         }

         if (suffix != null) {
            builder.append(suffix);
         }

         return hasRootStyle ? ((TextComponent.Builder)((TextComponent.Builder)Component.text().style(rootStyle)).append(builder)).build() : builder.build();
      }
   }

   // $FF: synthetic method
   JoinConfigurationImpl(BuilderImpl x0, Object x1) {
      this(x0);
   }

   static final class BuilderImpl implements JoinConfiguration.Builder {
      private ComponentLike prefix;
      private ComponentLike suffix;
      private ComponentLike separator;
      private ComponentLike lastSeparator;
      private ComponentLike lastSeparatorIfSerial;
      private Function<ComponentLike, Component> convertor;
      private Predicate<ComponentLike> predicate;
      private Style rootStyle;

      BuilderImpl() {
         this(JoinConfigurationImpl.NULL);
      }

      private BuilderImpl(final @NotNull JoinConfigurationImpl joinConfig) {
         this.separator = joinConfig.separator;
         this.lastSeparator = joinConfig.lastSeparator;
         this.prefix = joinConfig.prefix;
         this.suffix = joinConfig.suffix;
         this.convertor = joinConfig.convertor;
         this.lastSeparatorIfSerial = joinConfig.lastSeparatorIfSerial;
         this.predicate = joinConfig.predicate;
         this.rootStyle = joinConfig.rootStyle;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder prefix(final @Nullable ComponentLike prefix) {
         this.prefix = prefix;
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder suffix(final @Nullable ComponentLike suffix) {
         this.suffix = suffix;
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder separator(final @Nullable ComponentLike separator) {
         this.separator = separator;
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder lastSeparator(final @Nullable ComponentLike lastSeparator) {
         this.lastSeparator = lastSeparator;
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder lastSeparatorIfSerial(final @Nullable ComponentLike lastSeparatorIfSerial) {
         this.lastSeparatorIfSerial = lastSeparatorIfSerial;
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder convertor(final @NotNull Function<ComponentLike, Component> convertor) {
         this.convertor = (Function)Objects.requireNonNull(convertor, "convertor");
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder predicate(final @NotNull Predicate<ComponentLike> predicate) {
         this.predicate = (Predicate)Objects.requireNonNull(predicate, "predicate");
         return this;
      }

      @NotNull
      public JoinConfiguration.@NotNull Builder parentStyle(final @NotNull Style parentStyle) {
         this.rootStyle = (Style)Objects.requireNonNull(parentStyle, "rootStyle");
         return this;
      }

      public @NotNull JoinConfiguration build() {
         return new JoinConfigurationImpl(this);
      }

      // $FF: synthetic method
      BuilderImpl(JoinConfigurationImpl x0, Object x1) {
         this(x0);
      }
   }
}
